#ifndef LIBCLIENT_H
#define LIBCLIENT_H


#define LIBCLIENT_MAJOR 0
#define LIBCLIENT_MINOR 1
#define LIBCLIENT_PATCH 0


#include <stdbool.h>
#include <stdint.h>
#include "libkynput.h"

typedef enum {
    LIBCLIENT_LOG_LEVEL_ERROR,
    LIBCLIENT_LOG_LEVEL_WARN,
    LIBCLIENT_LOG_LEVEL_INFO,
    LIBCLIENT_LOG_LEVEL_DEBUG,
    LIBCLIENT_LOG_LEVEL_TRACE,
} libclient_LogLevel;

typedef enum {
    LIBCLIENT_VIDEO_CODEC_H264,
    LIBCLIENT_VIDEO_CODEC_HEVC,
    LIBCLIENT_VIDEO_CODEC_AV1,
} libclient_VideoCodec;

typedef enum {
    LIBCLIENT_EVENT_TYPE_CONTROL_PLANE_CONNECTED,
    LIBCLIENT_EVENT_TYPE_CONTROL_PLANE_CONNECT_FAILED,
    LIBCLIENT_EVENT_TYPE_DATA_PLANE_CONNECT_FAILED,
    LIBCLIENT_EVENT_TYPE_DATA_PLANE_CLOSED,
    LIBCLIENT_EVENT_TYPE_STOPPED,
    LIBCLIENT_EVENT_TYPE_DISPLAY_LIST_UPDATED,
    LIBCLIENT_EVENT_TYPE_INPUT_SERVICE_CONNECTED,
    LIBCLIENT_EVENT_TYPE_INPUT_SERVICE_DISCONNECTED,
    LIBCLIENT_EVENT_TYPE_STREAMING_STARTED,
    LIBCLIENT_EVENT_TYPE_STREAMING_START_FAILED,
    LIBCLIENT_EVENT_TYPE_STREAMING_STOPPED,
    LIBCLIENT_EVENT_TYPE_STREAMER_PIPELINE_EVENT,
    LIBCLIENT_EVENT_TYPE_AUDIO_VOLUME_CHANGED,
    LIBCLIENT_EVENT_TYPE_AUDIO_MUTE_CHANGED,
    LIBCLIENT_EVENT_TYPE_BITRATE_CHANGED,
} libclient_EventType;

typedef enum {
    LIBCLIENT_KYMUX_AUDIO_PROTOCOL_RELIABLE,
    LIBCLIENT_KYMUX_AUDIO_PROTOCOL_UNRELIABLE,
    LIBCLIENT_KYMUX_AUDIO_PROTOCOL_UNRELIABLE_FEC,
} libclient_KymuxAudioProtocol;

typedef enum {
    LIBCLIENT_KYMUX_VIDEO_PROTOCOL_RELIABLE,
    LIBCLIENT_KYMUX_VIDEO_PROTOCOL_GOP_STREAM,
    LIBCLIENT_KYMUX_VIDEO_PROTOCOL_UNRELIABLE,
    LIBCLIENT_KYMUX_VIDEO_PROTOCOL_UNRELIABLE_FEC,
} libclient_KymuxVideoProtocol;

typedef enum {
    LIBCLIENT_STOP_REASON_REQUESTED,
    LIBCLIENT_STOP_REASON_CONTROL_PLANE_CLOSED,
    LIBCLIENT_STOP_REASON_CONTROL_PLANE_IO_ERROR,
    LIBCLIENT_STOP_REASON_CONTROL_PLANE_TIMEOUT,
    LIBCLIENT_STOP_REASON_DECODE_ERROR,
} libclient_StopReason;

typedef enum {
    LIBCLIENT_STREAMER_PIPELINE_EVENT_TYPE_STOPPED,
    LIBCLIENT_STREAMER_PIPELINE_EVENT_TYPE_ERROR,
} libclient_StreamerPipelineEventType;

typedef enum {
    LIBCLIENT_STREAMER_PIPELINE_STAGE_PROCESS,
    LIBCLIENT_STREAMER_PIPELINE_STAGE_INITIALIZATION,
    LIBCLIENT_STREAMER_PIPELINE_STAGE_ACQUISITION,
    LIBCLIENT_STREAMER_PIPELINE_STAGE_ENCODER,
} libclient_StreamerPipelineStage;

typedef enum {
    LIBCLIENT_STREAMER_PIPELINE_TYPE_ENUMERATE_DISPLAYS,
    LIBCLIENT_STREAMER_PIPELINE_TYPE_AUDIO,
    LIBCLIENT_STREAMER_PIPELINE_TYPE_VIDEO,
} libclient_StreamerPipelineType;

typedef struct libclient_Event libclient_Event;

typedef struct libclient_AudioSessionConfig libclient_AudioSessionConfig;

typedef struct libclient_Client libclient_Client;

typedef struct libclient_ConnectConfig libclient_ConnectConfig;

typedef struct libclient_InputSessionConfig libclient_InputSessionConfig;

typedef struct libclient_StreamingProtocol libclient_StreamingProtocol;

typedef struct libclient_StreamingSessionConfig libclient_StreamingSessionConfig;

typedef struct libclient_VideoPlayerConfig libclient_VideoPlayerConfig;

typedef struct libclient_VideoSessionConfig libclient_VideoSessionConfig;

typedef struct libclient_WindowHandle libclient_WindowHandle;

typedef void (*libclient_LogCb)(libclient_LogLevel level,
                                const char *target,
                                const char *line,
                                void *userdata);

typedef struct {
    uint32_t id;
    int32_t height;
    int32_t width;
    int32_t x;
    int32_t y;
} libclient_Display;

typedef struct {
    void (*on_event)(const libclient_Event *event, void *userdata);
    void *userdata;
} libclient_EventHandler;

#ifdef __cplusplus
extern "C" {
#endif // __cplusplus

void libclient_init_crypto(void);

void libclient_set_log_cb(libclient_LogCb cb, void *userdata);

libclient_WindowHandle *libclient_windowshandle_from_x11(uint64_t handle);

libclient_WindowHandle *libclient_windowshandle_from_ns_view(uint64_t handle);

libclient_WindowHandle *libclient_windowshandle_from_hwnd(uint64_t hwnd);

int libclient_connect_config_default(const char *host, libclient_ConnectConfig **out_config);

int libclient_connect_config_set_tls_host(libclient_ConnectConfig *config, const char *tls_host);

int libclient_connect_config_set_tls_root_cert(libclient_ConnectConfig *config,
                                               const char *tls_root_cert);

void libclient_connect_config_set_streamer_port(libclient_ConnectConfig *config, uint16_t port);

void libclient_connect_config_unref(libclient_ConnectConfig *cfg);

libclient_StreamingProtocol *libclient_streaming_protocol_rtp(void);

libclient_StreamingProtocol *libclient_streaming_protocol_kymux(libclient_KymuxVideoProtocol video_protocol,
                                                                libclient_KymuxAudioProtocol audio_protocol);

void libclient_streaming_protocol_unref(libclient_StreamingProtocol *protocol);

void libclient_video_player_config_unref(libclient_VideoPlayerConfig *cfg);

int libclient_video_player_config_default(libclient_WindowHandle *window_handle,
                                          libclient_VideoPlayerConfig **out_config);

int libclient_video_session_config_default(uint32_t display_id,
                                           libclient_VideoPlayerConfig *player,
                                           libclient_VideoSessionConfig **out_config);

void libclient_video_player_config_set_software_decode(libclient_VideoPlayerConfig *config,
                                                       bool software_decode);

void libclient_video_session_config_unref(libclient_VideoSessionConfig *cfg);

void libclient_video_session_config_set_yuv444(libclient_VideoSessionConfig *config, bool yuv444);

void libclient_video_session_config_set_codec(libclient_VideoSessionConfig *config,
                                              libclient_VideoCodec codec);

void libclient_video_session_config_set_intra_refresh(libclient_VideoSessionConfig *config,
                                                      bool intra_refresh);

int libclient_audio_session_config_default(libclient_AudioSessionConfig **out_config);

void libclient_audio_session_config_set_audio_buffer(libclient_AudioSessionConfig *config,
                                                     uint32_t audio_buffer);

int libclient_input_session_config_default(libkynput_Pipeline *pipeline,
                                           libclient_InputSessionConfig **out_config);

void libclient_input_session_config_unref(libclient_InputSessionConfig *cfg);

void libclient_input_session_config_gamepad_force_xbox360(libclient_InputSessionConfig *cfg,
                                                          bool gamepad_force_xbox360);

int libclient_streaming_session_config_default(const libclient_StreamingProtocol *protocol,
                                               libclient_StreamingSessionConfig **out_config);

void libclient_streaming_session_config_unref(libclient_StreamingSessionConfig *cfg);

void libclient_streaming_session_config_set_bitrate(libclient_StreamingSessionConfig *config,
                                                    uint32_t bitrate);

uint32_t libclient_streaming_session_config_get_bitrate(libclient_StreamingSessionConfig *config);

void libclient_streaming_session_config_set_audio(libclient_StreamingSessionConfig *config,
                                                  libclient_AudioSessionConfig *audio);

void libclient_streaming_session_config_add_video(libclient_StreamingSessionConfig *config,
                                                  libclient_VideoSessionConfig *video);

void libclient_streaming_session_config_set_input(libclient_StreamingSessionConfig *config,
                                                  libclient_InputSessionConfig *input);

libclient_Client *libclient_client_new(libclient_ConnectConfig *connect_config);

int libclient_client_stop(libclient_Client *client);

int libclient_client_connect(libclient_Client *client);

int libclient_client_disconnect(libclient_Client *client);

int libclient_client_start_streaming(libclient_Client *client,
                                     libclient_StreamingSessionConfig *session_config);

int libclient_client_stop_streaming(libclient_Client *client);

int libclient_client_audio_set_volume(libclient_Client *client, float volume);

int libclient_client_audio_set_mute(libclient_Client *client, bool mute);

int libclient_client_set_bitrate(libclient_Client *client, uint32_t bitrate);

libclient_Client *libclient_client_ref(libclient_Client *client);

void libclient_client_unref(libclient_Client *client);

libclient_EventType libclient_event_get_type(const libclient_Event *event);

int libclient_event_unpack_control_plane_connected(const libclient_Event *event,
                                                   libclient_Display **out_displays,
                                                   uint32_t *out_len);

int libclient_event_unpack_stopped(const libclient_Event *event, libclient_StopReason *out_reason);

int libclient_event_unpack_display_list_updated(const libclient_Event *event,
                                                libclient_Display **out_displays,
                                                uint32_t *out_len);

void libclient_event_free_displays(libclient_Display *displays);

int libclient_event_unpack_streamer_pipeline_event(const libclient_Event *event,
                                                   libclient_StreamerPipelineType *out_type,
                                                   libclient_StreamerPipelineStage *out_stage,
                                                   libclient_StreamerPipelineEventType *out_event);

int libclient_event_unpack_audio_volume_changed(const libclient_Event *event, float *out_volume);

int libclient_event_unpack_audio_mute_changed(const libclient_Event *event, bool *out_mute);

int libclient_event_unpack_bitrate_changed(const libclient_Event *event, uint32_t *out_bitrate);

int libclient_client_set_event_handler(libclient_Client *client, libclient_EventHandler handler);

#ifdef __cplusplus
} // extern "C"
#endif // __cplusplus

#endif /* LIBCLIENT_H */
