#ifndef LIBKYNPUT_H
#define LIBKYNPUT_H


#define LIBKYNPUT_MAJOR 0
#define LIBKYNPUT_MINOR 1
#define LIBKYNPUT_PATCH 0


#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

/**
 * Control shape type.
 *
 * (x, y) is the position of the cursor in the Host screen.
 * Those values can be i32::MAX if the Host OS doesn't provide the position.
 */
typedef enum {
    LIBKYNPUT_CONTROL_EVENT_CLIENT_CONNECTED,
    LIBKYNPUT_CONTROL_EVENT_CLIENT_DISCONNECTED,
    LIBKYNPUT_CONTROL_EVENT_FLUSH_KEYS,
} libkynput_ControlEvent;

typedef enum {
    LIBKYNPUT_CURSOR_FORMAT_NONE,
    LIBKYNPUT_CURSOR_FORMAT_RAW,
    LIBKYNPUT_CURSOR_FORMAT_PNG,
} libkynput_CursorFormat;

/**
 * Gamepad axis.
 *
 * Axis value is contained in [`i16::MIN`]..[`i16::MAX`] range.
 *
 * [`GamepadAxis::TriggerLeft`] and [`GamepadAxis::TriggerRight`] are
 * special cases that can't have negative values.
 *
 * To be stored into a [`InputPacket`], it must be stored in a
 * [`GamepadData::Axis`] at first.
 */
typedef enum {
    LIBKYNPUT_GAMEPAD_AXIS_LEFT_X,
    LIBKYNPUT_GAMEPAD_AXIS_LEFT_Y,
    LIBKYNPUT_GAMEPAD_AXIS_RIGHT_X,
    LIBKYNPUT_GAMEPAD_AXIS_RIGHT_Y,
    LIBKYNPUT_GAMEPAD_AXIS_TRIGGER_LEFT,
    LIBKYNPUT_GAMEPAD_AXIS_TRIGGER_RIGHT,
} libkynput_GamepadAxis;

/**
 * Gamepad button.
 *
 * A button has only two stated: pressed or released.
 *
 * To be stored into a [`InputPacket`], it must be stored in a
 * [`GamepadData::Button`] at first.
 */
typedef enum {
    LIBKYNPUT_GAMEPAD_BUTTON_A,
    LIBKYNPUT_GAMEPAD_BUTTON_B,
    LIBKYNPUT_GAMEPAD_BUTTON_X,
    LIBKYNPUT_GAMEPAD_BUTTON_Y,
    LIBKYNPUT_GAMEPAD_BUTTON_BACK,
    LIBKYNPUT_GAMEPAD_BUTTON_GUIDE,
    LIBKYNPUT_GAMEPAD_BUTTON_START,
    LIBKYNPUT_GAMEPAD_BUTTON_LEFT_STICK,
    LIBKYNPUT_GAMEPAD_BUTTON_RIGHT_STICK,
    LIBKYNPUT_GAMEPAD_BUTTON_LEFT_SHOULDER,
    LIBKYNPUT_GAMEPAD_BUTTON_RIGHT_SHOULDER,
    LIBKYNPUT_GAMEPAD_BUTTON_D_PAD_UP,
    LIBKYNPUT_GAMEPAD_BUTTON_D_PAD_DOWN,
    LIBKYNPUT_GAMEPAD_BUTTON_D_PAD_LEFT,
    LIBKYNPUT_GAMEPAD_BUTTON_D_PAD_RIGHT,
} libkynput_GamepadButton;

typedef enum {
    LIBKYNPUT_NETWORK_STREAM_MSG_SERVING,
    LIBKYNPUT_NETWORK_STREAM_MSG_CLIENT_CONNECTED,
    LIBKYNPUT_NETWORK_STREAM_MSG_CLIENT_DISCONNECTED,
} libkynput_NetworkStreamMsg;

/**
 * Target of an [`InputPacket`]
 *
 * Every packet gets a target. This is required for [`InputRouter`](crate::router::InputRouter)
 * to decide where the packet should be forwarded.
 */
typedef enum {
    /**
     * The application that acquires the Keyboard/Mouse, and displays the Host's
     * video stream
     */
    LIBKYNPUT_INPUT_TARGET_CLIENT,
    /**
     * The application that records and stream a computer display. It also injects
     * the Keyboard/Mouse sent by the Client
     */
    LIBKYNPUT_INPUT_TARGET_HOST,
} libkynput_InputTarget;

/**
 * All supported kind of inputs.
 */
typedef enum {
    LIBKYNPUT_INPUT_TYPE_KEYBOARD,
    LIBKYNPUT_INPUT_TYPE_MOUSE_MOVE,
    LIBKYNPUT_INPUT_TYPE_MOUSE_POSITION,
    LIBKYNPUT_INPUT_TYPE_MOUSE_BUTTON,
    LIBKYNPUT_INPUT_TYPE_MOUSE_WHEEL,
    LIBKYNPUT_INPUT_TYPE_CURSOR,
    LIBKYNPUT_INPUT_TYPE_GAMEPAD_CONTROL,
    LIBKYNPUT_INPUT_TYPE_GAMEPAD_EVENT,
    LIBKYNPUT_INPUT_TYPE_CONTROL_EVENT,
} libkynput_InputType;

typedef enum {
    LIBKYNPUT_LOG_LEVEL_ERROR,
    LIBKYNPUT_LOG_LEVEL_WARN,
    LIBKYNPUT_LOG_LEVEL_INFO,
    LIBKYNPUT_LOG_LEVEL_DEBUG,
    LIBKYNPUT_LOG_LEVEL_TRACE,
} libkynput_LogLevel;

typedef enum {
    LIBKYNPUT_MOUSE_BUTTON_TYPE_LEFT,
    LIBKYNPUT_MOUSE_BUTTON_TYPE_MIDDLE,
    LIBKYNPUT_MOUSE_BUTTON_TYPE_RIGHT,
    LIBKYNPUT_MOUSE_BUTTON_TYPE_SIDE,
    LIBKYNPUT_MOUSE_BUTTON_TYPE_EXTRA,
} libkynput_MouseButtonType;

typedef enum {
    LIBKYNPUT_GAMEPAD_CONTROL_ID_PLUG,
    LIBKYNPUT_GAMEPAD_CONTROL_ID_UNPLUG,
} libkynput_GamepadControlId;

typedef enum {
    LIBKYNPUT_GAMEPAD_EVENT_ID_BUTTON,
    LIBKYNPUT_GAMEPAD_EVENT_ID_AXIS,
    LIBKYNPUT_GAMEPAD_EVENT_ID_RUMBLE,
} libkynput_GamepadEventId;

typedef enum {
    LIBKYNPUT_INPUT_TYPE_METADATA_ID_CURSOR,
} libkynput_InputTypeMetadataId;

/**
 * Cursor shape type.
 *
 * (x, y) is the position of the cursor in the Host screen.
 * Those values can be i32::MAX if the Host OS doesn't provide the position.
 */
typedef struct libkynput_Cursor libkynput_Cursor;

/**
 * Input packet
 *
 * Contains an [`InputTarget`] and a [`Payload`].
 */
typedef struct libkynput_Packet libkynput_Packet;

typedef struct libkynput_RouterHookControlResult libkynput_RouterHookControlResult;

/**
 * Keyboard key type.
 *
 * Scancode set 1 is used to represent the key identifier that should be
 * understood by all the OS implementations. An OS <=> Scancode set 1 mapping
 * should be done if required in the InputHandler implementations.
 *
 * Some sources to get more details about scancodes:
 * - https://blog.molecular-matters.com/2011/09/05/properly-handling-keyboard-input/
 * - https://docs.microsoft.com/en-us/windows/win32/api/winuser/ns-winuser-rawkeyboard#members
 * - https://download.microsoft.com/download/1/6/1/161ba512-40e2-4cc9-843a-923143f3456c/translate.pdf
 */
typedef struct libkynput_KeyboardKey libkynput_KeyboardKey;

/**
 * Mouse button type.
 */
typedef struct libkynput_MouseButton libkynput_MouseButton;

/**
 * Mouse relative move type.
 */
typedef struct libkynput_MouseMove libkynput_MouseMove;

/**
 * Mouse absolute move type.
 */
typedef struct libkynput_MousePosition libkynput_MousePosition;

/**
 * Mouse wheel move type.
 *
 * All OS (and even their different APIs) returns a different range of value to represent a
 * complete mouse wheel click.
 *
 * All producers must provide normalized values in [-1.0;1.0] range. 1.0 correspond to a single
 * physical mouse wheel click.
 *
 * All consumers must translate the normalized values into the OS API values.
 */
typedef struct libkynput_MouseWheel libkynput_MouseWheel;

typedef struct libkynput_Consumer libkynput_Consumer;

typedef struct libkynput_CursorData libkynput_CursorData;

typedef struct libkynput_Handler libkynput_Handler;

typedef struct libkynput_HostConfig libkynput_HostConfig;

typedef struct libkynput_InputTypeMetadata libkynput_InputTypeMetadata;

typedef struct libkynput_NetworkStream libkynput_NetworkStream;

typedef struct libkynput_Pipeline libkynput_Pipeline;

typedef struct libkynput_PipelineConfig libkynput_PipelineConfig;

typedef struct libkynput_RouterProcessingState libkynput_RouterProcessingState;

typedef void (*libkynput_LogCb)(libkynput_LogLevel level,
                                const char *target,
                                const char *line,
                                void *userdata);

typedef struct {
    bool process_event;
} libkynput_RouterHookPacketResult;

typedef struct {
    libkynput_RouterHookPacketResult (*on_packet)(const libkynput_Packet *packet, void *userdata);
    libkynput_RouterHookControlResult *(*on_control_event)(libkynput_ControlEvent evt,
                                                           void *userdata);
    void (*on_destroy)(void *userdata);
    void *userdata;
} libkynput_RouterHook;

typedef struct {
    void (*on_msg)(libkynput_NetworkStreamMsg msg, void *userdata);
    void (*on_destroy)(void *userdata);
    void *userdata;
} libkynput_NetworkStreamObserver;

typedef struct {
    void (*consume)(libkynput_Packet *packet, void *userdata);
    void (*destroy)(void *userdata);
    void *userdata;
} libkynput_ConsumerInterface;

typedef struct {
    void (*start)(void *userdata);
    void (*stop)(void *userdata);
    const libkynput_InputType *handled_types;
    size_t handled_types_count;
    void (*consume)(libkynput_Packet *packet, void *userdata);
    const libkynput_InputType *produced_types;
    size_t produced_types_count;
    libkynput_InputTypeMetadata *(*query_type_metadata)(libkynput_InputType type_, void *userdata);
    void (*destroy)(void *userdata);
    void *userdata;
} libkynput_HandlerInterface;

typedef uint32_t libkynput_GamepadId;

typedef struct {
    libkynput_GamepadId id;
} libkynput_GamepadControlIdPlug;

typedef struct {
    libkynput_GamepadId id;
} libkynput_GamepadControlIdUnplug;

typedef struct {
    libkynput_GamepadId id;
    libkynput_GamepadButton type_;
    bool pressed;
} libkynput_GamepadEventButton;

typedef struct {
    libkynput_GamepadId id;
    libkynput_GamepadAxis type_;
    int16_t value;
} libkynput_GamepadEventAxis;

typedef struct {
    libkynput_GamepadId id;
    uint16_t large_motor;
    uint16_t small_motor;
    int16_t duration;
} libkynput_GamepadEventRumble;

typedef struct {
    uint32_t id;
    int32_t height;
    int32_t width;
    int32_t x;
    int32_t y;
} libkynput_Display;

#ifdef __cplusplus
extern "C" {
#endif // __cplusplus

void libkynput_set_log_cb(libkynput_LogCb cb, void *userdata);

libkynput_RouterProcessingState *libkynput_router_processing_state_bool(bool value);

libkynput_RouterProcessingState *libkynput_router_processing_state_callback(bool (*cb)(libkynput_Packet *pkg,
                                                                                       void *userdata),
                                                                            void *userdata);

libkynput_Consumer *libkynput_consumer_ref(libkynput_Consumer *consumer);

void libkynput_consumer_unref(libkynput_Consumer *consumer);

int libkynput_consumer_consume(libkynput_Consumer *consumer, libkynput_Packet *pkt);

libkynput_Handler *libkynput_handler_ref(libkynput_Handler *instance);

void libkynput_handler_unref(libkynput_Handler *instance);

libkynput_Consumer *libkynput_handler_to_consumer(libkynput_Handler *instance);

libkynput_PipelineConfig *libkynput_get_dummy_client_config(void);

libkynput_PipelineConfig *libkynput_get_client_config(void);

libkynput_PipelineConfig *libkynput_get_host_config(const libkynput_HostConfig *host_config,
                                                    bool gamepad_force_xbox360);

libkynput_PipelineConfig *libkynput_get_dummy_host_config(const libkynput_HostConfig *host_config,
                                                          bool gamepad_force_xbox360);

void libkynput_pipeline_config_set_cursor_format(libkynput_PipelineConfig *config,
                                                 libkynput_CursorFormat cursor_format);

void libkynput_pipeline_config_unref(libkynput_PipelineConfig *config);

libkynput_Pipeline *libkynput_pipeline_new(libkynput_PipelineConfig *config);

libkynput_Pipeline *libkynput_pipeline_ref(libkynput_Pipeline *pipeline);

void libkynput_pipeline_unref(libkynput_Pipeline *pipeline);

int libkynput_pipeline_start(libkynput_Pipeline *pipeline);

int libkynput_pipeline_stop(libkynput_Pipeline *pipeline);

int libkynput_pipeline_update_host_config(libkynput_Pipeline *pipeline,
                                          const libkynput_HostConfig *host_config);

libkynput_Consumer *libkynput_pipeline_get_input_sink(libkynput_Pipeline *pipeline);

int libkynput_pipeline_plug_network(libkynput_Pipeline *pipeline, libkynput_Consumer *consumer);

int libkynput_pipeline_flush_keys(libkynput_Pipeline *pipeline);

bool libkynput_pipeline_produces_type(libkynput_Pipeline *pipeline, libkynput_InputType type_);

libkynput_InputTypeMetadata *libkynput_pipeline_query_type_metadata(libkynput_Pipeline *pipeline,
                                                                    libkynput_InputType type_);

int libkynput_pipeline_set_global_processing_state(libkynput_Pipeline *pipeline, bool value);

int libkynput_pipeline_set_type_processing_state(libkynput_Pipeline *pipeline,
                                                 libkynput_InputType type_,
                                                 libkynput_RouterProcessingState *state);

int libkynput_pipeline_plug_handler(libkynput_Pipeline *pipeline, libkynput_Handler *handler);

int libkynput_pipeline_set_hook(libkynput_Pipeline *pipeline, const libkynput_RouterHook *hook);

int libkynput_network_stream_plug_consumer(libkynput_NetworkStream *stream,
                                           libkynput_Consumer *consumer);

libkynput_NetworkStream *libkynput_network_stream_ref(libkynput_NetworkStream *stream);

void libkynput_network_stream_unref(libkynput_NetworkStream *stream);

libkynput_Consumer *libkynput_network_stream_ref_consumer(libkynput_NetworkStream *stream);

int libkynput_network_run(libkynput_NetworkStream *stream);

int libkynput_network_stop(libkynput_NetworkStream *stream);

libkynput_NetworkStream *libkynput_tcp_client_new(const char *host,
                                                  libkynput_NetworkStreamObserver *observer);

libkynput_NetworkStream *libkynput_tcp_server_new(libkynput_NetworkStreamObserver *observer);

libkynput_NetworkStream *libkynput_kymux_stream_new(libkynput_InputTarget target,
                                                    libkynput_NetworkStreamObserver *observer,
                                                    const char *uri);

libkynput_Packet *libkynput_packet_clone(const libkynput_Packet *packet);

void libkynput_packet_unref(libkynput_Packet *packet);

uint8_t libkynput_packet_get_type(const libkynput_Packet *packet);

bool libkynput_packet_eq(const libkynput_Packet *p1, const libkynput_Packet *p2);

char *libkynput_packet_to_string(const libkynput_Packet *packet);

void libkynput_free_string(char *s);

void libkynput_keyboard_key_unref(libkynput_KeyboardKey *key);

libkynput_KeyboardKey *libkynput_keyboard_key_from_packet(const libkynput_Packet *packet);

libkynput_Packet *libkynput_keyboard_key_to_packet(libkynput_InputTarget target,
                                                   uint16_t scancode,
                                                   bool pressed);

bool libkynput_keyboard_key_pressed(const libkynput_KeyboardKey *key);

uint16_t libkynput_keyboard_key_scancode(const libkynput_KeyboardKey *key);

void libkynput_mouse_button_unref(libkynput_MouseButton *mouse_button);

libkynput_MouseButton *libkynput_mouse_button_from_packet(const libkynput_Packet *packet);

libkynput_Packet *libkynput_mouse_button_to_packet(libkynput_InputTarget target,
                                                   libkynput_MouseButtonType button,
                                                   bool pressed);

libkynput_MouseButtonType libkynput_mouse_button_button(const libkynput_MouseButton *mouse_button);

bool libkynput_mouse_button_pressed(const libkynput_MouseButton *mouse_button);

void libkynput_mouse_move_unref(libkynput_MouseMove *mouse_move);

libkynput_MouseMove *libkynput_mouse_move_from_packet(const libkynput_Packet *packet);

libkynput_Packet *libkynput_mouse_move_to_packet(libkynput_InputTarget target,
                                                 int16_t dx,
                                                 int16_t dy);

int16_t libkynput_mouse_move_dx(const libkynput_MouseMove *mouse_move);

int16_t libkynput_mouse_move_dy(const libkynput_MouseMove *mouse_move);

void libkynput_mouse_position_unref(libkynput_MousePosition *position);

libkynput_Packet *libkynput_mouse_position_to_packet(libkynput_InputTarget target,
                                                     uint32_t display_id,
                                                     int16_t x,
                                                     int16_t y);

libkynput_MousePosition *libkynput_mouse_position_from_packet(const libkynput_Packet *packet);

uint32_t libkynput_mouse_position_display_id(const libkynput_MousePosition *position);

int16_t libkynput_mouse_position_x(const libkynput_MousePosition *position);

int16_t libkynput_mouse_position_y(const libkynput_MousePosition *position);

void libkynput_mouse_wheel_unref(libkynput_MouseWheel *mouse_wheel);

libkynput_Packet *libkynput_mouse_wheel_to_packet(libkynput_InputTarget target, float dx, float dy);

libkynput_MouseWheel *libkynput_mouse_wheel_from_packet(const libkynput_Packet *packet);

float libkynput_mouse_wheel_dx(const libkynput_MouseWheel *mouse_wheel);

float libkynput_mouse_wheel_dy(const libkynput_MouseWheel *mouse_wheel);

void libkynput_cursor_unref(libkynput_Cursor *cursor);

libkynput_CursorData *libkynput_cursor_data_from_raw(const uint32_t *data, uint32_t size);

libkynput_CursorData *libkynput_cursor_data_from_png(const uint8_t *data, uint32_t size);

void libkynput_cursor_data_unref(libkynput_CursorData *data);

libkynput_Packet *libkynput_cursor_to_packet(libkynput_InputTarget target,
                                             uint32_t display_id,
                                             bool visible,
                                             int32_t x,
                                             int32_t y,
                                             uint32_t xhot,
                                             uint32_t yhot,
                                             uint32_t width,
                                             uint32_t height,
                                             const libkynput_CursorData *data);

libkynput_Cursor *libkynput_cursor_from_packet(const libkynput_Packet *packet);

uint32_t libkynput_cursor_display_id(const libkynput_Cursor *cursor);

bool libkynput_cursor_visible(const libkynput_Cursor *cursor);

int32_t libkynput_cursor_x(const libkynput_Cursor *cursor);

int32_t libkynput_cursor_y(const libkynput_Cursor *cursor);

uint32_t libkynput_cursor_xhot(const libkynput_Cursor *cursor);

uint32_t libkynput_cursor_yhot(const libkynput_Cursor *cursor);

uint32_t libkynput_cursor_width(const libkynput_Cursor *cursor);

uint32_t libkynput_cursor_height(const libkynput_Cursor *cursor);

libkynput_CursorFormat libkynput_cursor_format(const libkynput_Cursor *cursor);

const uint32_t *libkynput_cursor_raw_data(const libkynput_Cursor *cursor, uint32_t *size);

const uint8_t *libkynput_cursor_png_data(const libkynput_Cursor *cursor, uint32_t *size);

libkynput_Consumer *libkynput_consumer_new(libkynput_ConsumerInterface consumer);

libkynput_Handler *libkynput_handler_new(libkynput_HandlerInterface handler);

bool libkynput_routerhook_packet_result_eq(const libkynput_RouterHookPacketResult *a,
                                           const libkynput_RouterHookPacketResult *b);

libkynput_RouterHookControlResult *libkynput_router_hook_control_result_new(libkynput_Packet **packets,
                                                                            size_t count);

bool libkynput_gamepad_control_id_from_packet(const libkynput_Packet *packet,
                                              libkynput_GamepadControlId *out_id);

bool libkynput_gamepad_plug_from_packet(const libkynput_Packet *packet,
                                        libkynput_GamepadControlIdPlug *out_control);

libkynput_Packet *libkynput_gamepad_plug_to_packet(libkynput_InputTarget target,
                                                   libkynput_GamepadId id,
                                                   const uint16_t *vendor_id,
                                                   const uint16_t *product_id);

bool libkynput_gamepad_unplug_from_packet(const libkynput_Packet *packet,
                                          libkynput_GamepadControlIdUnplug *out_control);

libkynput_Packet *libkynput_gamepad_unplug_to_packet(libkynput_InputTarget target,
                                                     libkynput_GamepadId id);

bool libkynput_gamepad_event_id_from_packet(const libkynput_Packet *packet,
                                            libkynput_GamepadEventId *out_id);

bool libkynput_gamepad_button_from_packet(const libkynput_Packet *packet,
                                          libkynput_GamepadEventButton *out_button);

libkynput_Packet *libkynput_gamepad_button_to_packet(libkynput_InputTarget target,
                                                     const libkynput_GamepadEventButton *button);

bool libkynput_gamepad_axis_from_packet(const libkynput_Packet *packet,
                                        libkynput_GamepadEventAxis *out_axis);

libkynput_Packet *libkynput_gamepad_axis_to_packet(libkynput_InputTarget target,
                                                   const libkynput_GamepadEventAxis *axis);

bool libkynput_gamepad_rumble_from_packet(const libkynput_Packet *packet,
                                          libkynput_GamepadEventRumble *out_rumble);

libkynput_HostConfig *libkynput_host_config_new(const libkynput_Display *display_list,
                                                size_t count);

void libkynput_host_config_unref(libkynput_HostConfig *host_config);

libkynput_InputTypeMetadataId libkynput_input_type_metadata_id(const libkynput_InputTypeMetadata *metadata);

void libkynput_input_type_metadata_unref(libkynput_InputTypeMetadata *metadata);

libkynput_InputTypeMetadata *libkynput_input_type_metadata_cursor_pack(libkynput_CursorFormat format);

bool libkynput_input_type_metadata_cursor_unpack(const libkynput_InputTypeMetadata *metadata,
                                                 libkynput_CursorFormat *out_format);

libkynput_Packet *libkynput_control_event_to_packet(libkynput_InputTarget target,
                                                    libkynput_ControlEvent event);

#ifdef __cplusplus
} // extern "C"
#endif // __cplusplus

#endif /* LIBKYNPUT_H */
